@@ -1,5 +1,6 @@
 #include <ctype.h>
 #include <errno.h>
+#include <libgen.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -9,14 +10,62 @@
 #include "config.h"
 #include "log.h"
 #include "nelem.h"
+#include "unicode.h"
 #include "xmalloc.h"
 
 /* Maximum number of config file errors before we give up */
 #define MAX_ERRORS 5
 
+/* Maximum inclusion recursion depth before we give up */
+#define MAX_RECURSION 32
+
 /* Anyone with a 10M config file is doing something very wrong */
 #define MAX_CONFIG_SIZE (10*1024*1024)
 
+/* Convenience macros for anchor combinations */
+#define ANCHOR_TOP_LEFT (\
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT \
+		)
+#define ANCHOR_TOP (\
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT \
+		)
+#define ANCHOR_TOP_RIGHT (\
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT \
+		)
+#define ANCHOR_RIGHT (\
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM \
+		)
+#define ANCHOR_BOTTOM_RIGHT (\
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT \
+		)
+#define ANCHOR_BOTTOM (\
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT \
+		)
+#define ANCHOR_BOTTOM_LEFT (\
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT \
+		)
+#define ANCHOR_LEFT (\
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM \
+		)
+#define ANCHOR_CENTER (\
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT \
+		| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT \
+		)
+
 struct uint32_percent {
 	uint32_t value;
 	bool percent;
@@ -29,6 +78,7 @@ static uint32_t fixup_percentage(uint32_t value, uint32_t base, bool is_percent,
 
 static uint32_t parse_anchor(const char *filename, size_t lineno, const char *str, bool *err);
 static bool parse_bool(const char *filename, size_t lineno, const char *str, bool *err);
+static uint32_t parse_char(const char *filename, size_t lineno, const char *str, bool *err);
 static struct color parse_color(const char *filename, size_t lineno, const char *str, bool *err);
 static uint32_t parse_uint32(const char *filename, size_t lineno, const char *str, bool *err);
 static int32_t parse_int32(const char *filename, size_t lineno, const char *str, bool *err);
@@ -63,6 +113,17 @@ void config_load(struct tofi *tofi, const char *filename)
 		}
 		filename = default_filename;
 	}
+	/*
+	 * Track and limit recursion depth, so we don't overflow the stack if
+	 * a config file loop is created.
+	 */
+	static uint8_t recursion_depth = 0;
+	recursion_depth++;
+	if (recursion_depth > MAX_RECURSION) {
+		log_error("Refusing to load %s, recursion too deep (>%u layers).\n", filename, MAX_RECURSION);
+		recursion_depth--;
+		return;
+	}
 	char *config;
 	FILE *fp = fopen(filename, "rb");
 	if (!fp) {
@@ -205,6 +266,7 @@ CLEANUP_FILENAME:
 	if (default_filename) {
 		free(default_filename);
 	}
+	recursion_depth--;
 }
 
 char *strip(const char *str)
@@ -238,7 +300,20 @@ bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const
 {
 	bool err = false;
 	struct uint32_percent percent;
-	if (strcasecmp(option, "anchor") == 0) {
+	if (strcasecmp(option, "include") == 0) {
+		if (value[0] == '/') {
+			config_load(tofi, value);
+		} else {
+			char *tmp = xstrdup(filename);
+			char *dir = dirname(tmp);
+			size_t len = strlen(dir) + 1 + strlen(value) + 1;
+			char *config = xcalloc(len, 1);
+			snprintf(config, len, "%s/%s", dir, value);
+			config_load(tofi, config);
+			free(config);
+			free(tmp);
+		}
+	} else if (strcasecmp(option, "anchor") == 0) {
 		tofi->anchor = parse_anchor(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "background-color") == 0) {
 		tofi->window.entry.background_color = parse_color(filename, lineno, value, &err);
@@ -278,6 +353,14 @@ bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const
 		tofi->window.entry.selection_background_padding = parse_int32(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "selection-background") == 0) {
 		tofi->window.entry.selection_background_color = parse_color(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "exclusive-zone") == 0) {
+		if (strcmp(value, "-1") == 0) {
+			tofi->window.exclusive_zone = -1;
+		} else {
+			percent = parse_uint32_percent(filename, lineno, value, &err);
+			tofi->window.exclusive_zone = percent.value;
+			tofi->window.exclusive_zone_is_percent = percent.percent;
+		}
 	} else if (strcasecmp(option, "width") == 0) {
 		percent = parse_uint32_percent(filename, lineno, value, &err);
 		tofi->window.width = percent.value;
@@ -328,13 +411,25 @@ bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const
 		tofi->fuzzy_match = parse_bool(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "require-match") == 0) {
 		tofi->require_match = parse_bool(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "hide-input") == 0) {
+		tofi->window.entry.hide_input = parse_bool(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "hidden-character") == 0) {
+		uint32_t ch = parse_char(filename, lineno, value, &err);
+		if (!err) {
+			tofi->window.entry.hidden_character_utf8_length = 
+				utf32_to_utf8(ch, tofi->window.entry.hidden_character_utf8);
+		}
 	} else if (strcasecmp(option, "drun-launch") == 0) {
 		tofi->drun_launch = parse_bool(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "drun-print-exec") == 0) {
 		log_warning("drun-print-exec is deprecated, as it is now always true.\n"
 				"           This option may be removed in a future version of tofi.\n");
+	} else if (strcasecmp(option, "terminal") == 0) {
+		snprintf(tofi->default_terminal, N_ELEM(tofi->default_terminal), "%s", value);
 	} else if (strcasecmp(option, "hint-font") == 0) {
 		tofi->window.entry.harfbuzz.disable_hinting = !parse_bool(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "multi-instance") == 0) {
+		tofi->multiple_instance = parse_bool(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "late-keyboard-init") == 0) {
 		tofi->late_keyboard_init = parse_bool(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "output") == 0) {
@@ -443,6 +538,39 @@ void config_fixup_values(struct tofi *tofi)
 			tofi->window.entry.padding_right_is_percent,
 			tofi->window.scale,
 			tofi->use_scale);
+
+	/* Don't attempt percentage handling if exclusive_zone is set to -1. */
+	if (tofi->window.exclusive_zone > 0) {
+		/* Exclusive zone base depends on anchor. */
+		switch (tofi->anchor) {
+			case ANCHOR_TOP:
+			case ANCHOR_BOTTOM:
+				tofi->window.exclusive_zone = fixup_percentage(
+						tofi->window.exclusive_zone,
+						tofi->output_height,
+						tofi->window.exclusive_zone_is_percent,
+						tofi->window.scale,
+						tofi->use_scale);
+				break;
+			case ANCHOR_LEFT:
+			case ANCHOR_RIGHT:
+				tofi->window.exclusive_zone = fixup_percentage(
+						tofi->window.exclusive_zone,
+						tofi->output_width,
+						tofi->window.exclusive_zone_is_percent,
+						tofi->window.scale,
+						tofi->use_scale);
+				break;
+			default:
+				/*
+				 * Exclusive zone >0 is meaningless for other anchor
+				 * positions.
+				 */
+				tofi->window.exclusive_zone =
+					MIN(tofi->window.exclusive_zone, 0);
+				break;
+		}
+	}
 }
 
 char *get_config_path()
@@ -478,49 +606,45 @@ bool parse_bool(const char *filename, size_t lineno, const char *str, bool *err)
 	return false;
 }
 
+uint32_t parse_char(const char *filename, size_t lineno, const char *str, bool *err)
+{
+	char *tmp = utf8_compose(str);
+	uint32_t ch = utf8_to_utf32(tmp);
+	if (*utf8_next_char(tmp) != '\0') {
+		PARSE_ERROR(filename, lineno, "Failed to parse \"%s\" as a character.\n", str);
+	}
+	free(tmp);
+	return ch;
+}
+
 uint32_t parse_anchor(const char *filename, size_t lineno, const char *str, bool *err)
 {
 	if(strcasecmp(str, "top-left") == 0) {
-		return ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT;
+		return ANCHOR_TOP_LEFT;
 	}
 	if (strcasecmp(str, "top") == 0) {
-		return ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
+		return ANCHOR_TOP;
 	}
 	if (strcasecmp(str, "top-right") == 0) {
-		return ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
+		return ANCHOR_TOP_RIGHT;
 	}
 	if (strcasecmp(str, "right") == 0) {
-		return ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM;
+		return ANCHOR_RIGHT;
 	}
 	if (strcasecmp(str, "bottom-right") == 0) {
-		return ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
+		return ANCHOR_BOTTOM_RIGHT;
 	}
 	if (strcasecmp(str, "bottom") == 0) {
-		return ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
+		return ANCHOR_BOTTOM;
 	}
 	if (strcasecmp(str, "bottom-left") == 0) {
-		return ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT;
+		return ANCHOR_BOTTOM_LEFT;
 	}
 	if (strcasecmp(str, "left") == 0) {
-		return ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM;
+		return ANCHOR_LEFT;
 	}
 	if (strcasecmp(str, "center") == 0) {
-		return ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT
-			| ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
+		return ANCHOR_CENTER;
 	}
 	PARSE_ERROR(filename, lineno, "Invalid anchor \"%s\".\n", str);
 	*err = true;
