@@ -12,8 +12,6 @@
 #include <unistd.h>
 #include <wayland-client.h>
 #include <wayland-util.h>
-#include <wchar.h>
-#include <wctype.h>
 #include <xkbcommon/xkbcommon.h>
 #include "tofi.h"
 #include "compgen.h"
@@ -21,10 +19,13 @@
 #include "config.h"
 #include "entry.h"
 #include "image.h"
+#include "input.h"
 #include "log.h"
 #include "nelem.h"
+#include "lock.h"
 #include "shm.h"
 #include "string_vec.h"
+#include "string_vec.h"
 #include "xmalloc.h"
 
 #undef MAX
@@ -142,151 +143,6 @@ static void wl_keyboard_leave(
 	/* Deliberately left blank */
 }
 
-static void handle_keypress(struct tofi *tofi, xkb_keycode_t keycode)
-{
-	if (tofi->xkb_state == NULL) {
-		return;
-	}
-	xkb_keysym_t sym = xkb_state_key_get_one_sym(tofi->xkb_state, keycode);
-
-	struct entry *entry = &tofi->window.entry;
-	char buf[5]; /* 4 UTF-8 bytes plus null terminator. */
-	int len = xkb_state_key_get_utf8(
-			tofi->xkb_state,
-			keycode,
-			buf,
-			sizeof(buf));
-	wchar_t ch;
-	mbtowc(&ch, buf, sizeof(buf));
-	if (len > 0 && iswprint(ch)) {
-		if (entry->input_length < N_ELEM(entry->input) - 1) {
-			entry->input[entry->input_length] = ch;
-			entry->input_length++;
-			entry->input[entry->input_length] = L'\0';
-			memcpy(&entry->input_mb[entry->input_mb_length],
-					buf,
-					N_ELEM(buf));
-			entry->input_mb_length += len;
-			if (entry->drun) {
-				struct string_vec results = desktop_vec_filter(&entry->apps, entry->input_mb, tofi->fuzzy_match);
-				string_vec_destroy(&entry->results);
-				entry->results = results;
-			} else {
-				struct string_vec tmp = entry->results;
-				entry->results = string_vec_filter(&entry->results, entry->input_mb, tofi->fuzzy_match);
-				string_vec_destroy(&tmp);
-			}
-		}
-	} else if (entry->input_length > 0
-			&& (sym == XKB_KEY_BackSpace
-				|| (sym == XKB_KEY_w
-					&& xkb_state_mod_name_is_active(
-						tofi->xkb_state,
-						XKB_MOD_NAME_CTRL,
-						XKB_STATE_MODS_EFFECTIVE))))
-	{
-		if (sym == XKB_KEY_BackSpace) {
-			entry->input_length--;
-			entry->input[entry->input_length] = L'\0';
-		} else {
-			while (entry->input_length > 0 && iswspace(entry->input[entry->input_length - 1])) {
-				entry->input_length--;
-			}
-			while (entry->input_length > 0 && !iswspace(entry->input[entry->input_length - 1])) {
-				entry->input_length--;
-			}
-			entry->input[entry->input_length] = L'\0';
-		}
-		const wchar_t *src = entry->input;
-		size_t siz = wcsrtombs(
-				entry->input_mb,
-				&src,
-				N_ELEM(entry->input_mb),
-				NULL);
-		entry->input_mb_length = siz;
-		string_vec_destroy(&entry->results);
-		if (entry->drun) {
-			entry->results = desktop_vec_filter(&entry->apps, entry->input_mb, tofi->fuzzy_match);
-		} else {
-			entry->results = string_vec_filter(&entry->commands, entry->input_mb, tofi->fuzzy_match);
-		}
-	} else if (sym == XKB_KEY_u
-			&& xkb_state_mod_name_is_active(
-				tofi->xkb_state,
-				XKB_MOD_NAME_CTRL,
-				XKB_STATE_MODS_EFFECTIVE)
-		  )
-	{
-		entry->input_length = 0;
-		entry->input[0] = L'\0';
-		entry->input_mb_length = 0;
-		entry->input_mb[0] = '\0';
-		string_vec_destroy(&entry->results);
-		if (entry->drun) {
-			entry->results = desktop_vec_filter(&entry->apps, entry->input_mb, tofi->fuzzy_match);
-		} else {
-			entry->results = string_vec_filter(&entry->commands, entry->input_mb, tofi->fuzzy_match);
-		}
-	} else if (sym == XKB_KEY_Escape
-			|| (sym == XKB_KEY_c
-				&& xkb_state_mod_name_is_active(
-					tofi->xkb_state,
-					XKB_MOD_NAME_CTRL,
-					XKB_STATE_MODS_EFFECTIVE)
-			   )
-		  )
-	{
-		tofi->closed = true;
-		return;
-	} else if (sym == XKB_KEY_Return || sym == XKB_KEY_KP_Enter) {
-		tofi->submit = true;
-		return;
-	}
-
-	uint32_t nsel = MAX(MIN(entry->num_results_drawn, entry->results.count), 1);
-	if (sym == XKB_KEY_Up || sym == XKB_KEY_Left
-			|| (sym == XKB_KEY_k
-				&& xkb_state_mod_name_is_active(
-					tofi->xkb_state,
-					XKB_MOD_NAME_CTRL,
-					XKB_STATE_MODS_EFFECTIVE)
-			   )
-			) {
-		if (entry->selection > 0) {
-			entry->selection--;
-		} else {
-			if (entry->first_result > nsel) {
-				entry->first_result -= entry->last_num_results_drawn;
-				entry->selection = entry->last_num_results_drawn - 1;
-			} else if (entry->first_result > 0) {
-				entry->selection = entry->first_result - 1;
-				entry->first_result = 0;
-			}
-		}
-	} else if (sym == XKB_KEY_Down || sym == XKB_KEY_Right || sym == XKB_KEY_Tab
-			|| (sym == XKB_KEY_j
-				&& xkb_state_mod_name_is_active(
-					tofi->xkb_state,
-					XKB_MOD_NAME_CTRL,
-					XKB_STATE_MODS_EFFECTIVE)
-			   )
-			) {
-		entry->selection++;
-		if (entry->selection >= nsel) {
-			entry->selection -= nsel;
-			entry->first_result += nsel;
-			entry->first_result %= entry->results.count;
-			entry->last_num_results_drawn = entry->num_results_drawn;
-		}
-	} else {
-		entry->selection = 0;
-		entry->first_result = 0;
-	}
-	entry_update(&tofi->window.entry);
-	tofi->window.surface.redraw = true;
-	
-}
-
 static void wl_keyboard_key(
 		void *data,
 		struct wl_keyboard *wl_keyboard,
@@ -317,7 +173,7 @@ static void wl_keyboard_key(
 		tofi->repeat.keycode = keycode;
 		tofi->repeat.next = gettime_ms() + tofi->repeat.delay;
 	}
-	handle_keypress(tofi, keycode);
+	input_handle_keypress(tofi, keycode);
 }
 
 static void wl_keyboard_modifiers(
@@ -775,6 +631,7 @@ static void usage()
 "\n"
 "  -h, --help                           Print this message and exit.\n"
 "  -c, --config <path>                  Specify a config file.\n"
+"      --include <path>                 Include an additional config file.\n"
 "      --font <name|path>               Font to use.\n"
 "      --font-size <pt>                 Point size of text.\n"
 "      --background-color <color>       Color of the background.\n"
@@ -800,6 +657,7 @@ static void usage()
 "      --output <name>                  Name of output to display window on.\n"
 "      --scale <true|false>             Follow the output's scale factor.\n"
 "      --anchor <position>              Location on screen to anchor window.\n"
+"      --exclusive-zone <-1|px|%>       Exclusive zone size, or -1 for none.\n"
 "      --margin-top <px|%>              Offset from top of screen.\n"
 "      --margin-bottom <px|%>           Offset from bottom of screen.\n"
 "      --margin-left <px|%>             Offset from left of screen.\n"
@@ -813,11 +671,16 @@ static void usage()
 "      --history <true|false>           Sort results by number of usages.\n"
 "      --fuzzy-match <true|false>       Use fuzzy matching for searching.\n"
 "      --require-match <true|false>     Require a match for selection.\n"
+"      --hide-input <true|false>        Hide sensitive input such as passwords.\n"
+"      --hidden-character <char>        Replacement character for hidden input.\n"
 "      --drun-launch <true|false>       Launch apps directly in drun mode.\n"
 "      --drun-print-exec <true|false>   Print a command line in drun mode.\n"
 "                                       This is now always the case,\n"
 "                                       and this option is deprecated.\n"
+"      --terminal <command>             Terminal to use for command line\n"
+"                                       programs in drun mode.\n"
 "      --hint-font <true|false>         Perform font hinting.\n"
+"      --multi-instance <true|false>    Allow multiple tofi instances at once.\n"
 "      --late-keyboard-init             (EXPERIMENTAL) Delay keyboard\n"
 "                                       initialisation until after the first\n"
 "                                       draw to screen.\n"
@@ -828,7 +691,9 @@ static void usage()
 const struct option long_options[] = {
 	{"help", no_argument, NULL, 'h'},
 	{"config", required_argument, NULL, 'c'},
+	{"include", required_argument, NULL, 0},
 	{"anchor", required_argument, NULL, 0},
+	{"exclusive-zone", required_argument, NULL, 0},
 	{"background-color", required_argument, NULL, 0},
 	{"corner-radius", required_argument, NULL, 0},
 	{"font", required_argument, NULL, 0},
@@ -862,9 +727,13 @@ const struct option long_options[] = {
 	{"history", required_argument, NULL, 0},
 	{"fuzzy-match", required_argument, NULL, 0},
 	{"require-match", required_argument, NULL, 0},
+	{"hide-input", required_argument, NULL, 0},
+	{"hidden-character", required_argument, NULL, 0},
 	{"drun-launch", required_argument, NULL, 0},
 	{"drun-print-exec", required_argument, NULL, 0},
+	{"terminal", required_argument, NULL, 0},
 	{"hint-font", required_argument, NULL, 0},
+	{"multi-instance", required_argument, NULL, 0},
 	{"output", required_argument, NULL, 0},
 	{"scale", required_argument, NULL, 0},
 	{"late-keyboard-init", optional_argument, NULL, 'k'},
@@ -948,7 +817,7 @@ static bool do_submit(struct tofi *tofi)
 		if (tofi->require_match || entry->drun) {
 			return false;
 		} else {
-			printf("%ls\n", entry->input);
+			printf("%s\n", entry->input_utf8);
 			return true;
 		}
 	}
@@ -970,7 +839,7 @@ static bool do_submit(struct tofi *tofi)
 		if (tofi->drun_launch) {
 			drun_launch(res);
 		} else {
-			drun_print(res);
+			drun_print(res, tofi->default_terminal);
 		}
 	} else {
 		printf("%s\n", res);
@@ -1001,10 +870,12 @@ int main(int argc, char *argv[])
 			.scale = 1,
 			.width = 1280,
 			.height = 720,
+			.exclusive_zone = -1,
 			.entry = {
 				.font_name = "Sans",
 				.font_size = 24,
 				.prompt_text = "run: ",
+				.hidden_character_utf8 = u8"*",
 				.padding_top = 8,
 				.padding_bottom = 8,
 				.padding_left = 8,
@@ -1027,9 +898,21 @@ int main(int argc, char *argv[])
 		.use_scale = true,
 	};
 	wl_list_init(&tofi.output_list);
+	if (getenv("TERMINAL") != NULL) {
+		snprintf(
+				tofi.default_terminal,
+				N_ELEM(tofi.default_terminal),
+				"%s",
+				getenv("TERMINAL"));
+	}
 
 	parse_args(&tofi, argc, argv);
 
+	if (!tofi.multiple_instance && lock_check()) {
+		log_error("Another instance of tofi is already running.\n");
+		exit(EXIT_FAILURE);
+	}
+
 	/*
 	 * Initial Wayland & XKB setup.
 	 * The first thing to do is connect a listener to the global registry,
@@ -1107,6 +990,14 @@ int main(int argc, char *argv[])
 					NULL,
 					ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND,
 					"dummy");
+		/*
+		 * Workaround for Hyprland, where if this is not set the dummy
+		 * surface never enters an output for some reason.
+		 */
+		zwlr_layer_surface_v1_set_keyboard_interactivity(
+				zwlr_layer_surface,
+				ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE
+				);
 		zwlr_layer_surface_v1_add_listener(
 				zwlr_layer_surface,
 				&dummy_layer_surface_listener,
@@ -1276,7 +1167,7 @@ int main(int argc, char *argv[])
 			tofi.zwlr_layer_shell,
 			tofi.window.surface.wl_surface,
 			wl_output,
-			ZWLR_LAYER_SHELL_V1_LAYER_TOP,
+			ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY,
 			"launcher");
 	zwlr_layer_surface_v1_set_keyboard_interactivity(
 			tofi.window.zwlr_layer_surface,
@@ -1291,7 +1182,7 @@ int main(int argc, char *argv[])
 			tofi.anchor);
 	zwlr_layer_surface_v1_set_exclusive_zone(
 			tofi.window.zwlr_layer_surface,
-			-1);
+			tofi.window.exclusive_zone);
 	zwlr_layer_surface_v1_set_size(
 			tofi.window.zwlr_layer_surface,
 			tofi.window.width / tofi.window.scale,
@@ -1429,7 +1320,7 @@ int main(int argc, char *argv[])
 			if (tofi.repeat.active) {
 				int64_t wait = (int64_t)tofi.repeat.next - (int64_t)gettime_ms();
 				if (wait <= 0) {
-					handle_keypress(&tofi, tofi.repeat.keycode);
+					input_handle_keypress(&tofi, tofi.repeat.keycode);
 					tofi.repeat.next += 1000 / tofi.repeat.rate;
 				}
 			}
@@ -1445,6 +1336,7 @@ int main(int argc, char *argv[])
 		wl_display_dispatch_pending(tofi.wl_display);
 
 		if (tofi.window.surface.redraw) {
+			entry_update(&tofi.window.entry);
 			surface_draw(&tofi.window.surface);
 			tofi.window.surface.redraw = false;
 		}
