@@ -11,6 +11,7 @@
 #include "wlr-layer-shell-unstable-v1.h"
 
 #define MAX_OUTPUT_NAME_LEN 256
+#define MAX_TERMINAL_NAME_LEN 256
 
 struct output_list_element {
 	struct wl_list link;
@@ -56,12 +57,14 @@ struct tofi {
 		uint32_t height;
 		uint32_t scale;
 		int32_t transform;
+		int32_t exclusive_zone;
 		int32_t margin_top;
 		int32_t margin_bottom;
 		int32_t margin_left;
 		int32_t margin_right;
 		bool width_is_percent;
 		bool height_is_percent;
+		bool exclusive_zone_is_percent;
 		bool margin_top_is_percent;
 		bool margin_bottom_is_percent;
 		bool margin_left_is_percent;
@@ -85,7 +88,9 @@ struct tofi {
 	bool drun_print_exec;
 	bool fuzzy_match;
 	bool require_match;
+	bool multiple_instance;
 	char target_output_name[MAX_OUTPUT_NAME_LEN];
+	char default_terminal[MAX_TERMINAL_NAME_LEN];
 };
 
 #endif /* TOFI_H */
